/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 10/07/23.
//

#ifndef FGGL_COMPONENTS_CORE_SERVICES_FACTORY_HPP
#define FGGL_COMPONENTS_CORE_SERVICES_FACTORY_HPP

#include <algorithm>
#include <functional>
#include <optional>
#include <cassert>
#include <memory>

#include "registry.hpp"


namespace fggl::services {

	class Generator;

	struct ServiceBinding {
		public:
			virtual ~ServiceBinding() = default;
			virtual std::shared_ptr<void> provide() = 0;
	};

	template<ServiceAPI T>
	struct ServiceBindingT;

	class Generator {
		public:
			Generator() = default;

			// no copy
			Generator(Generator&) = delete;
			void operator=(Generator&) = delete;

			template<ServiceAPI api>
			Service<api> getLazy() {
				assert( m_factories.find(api::SERVICE_ID) != m_factories.end() );
				if ( m_registry.contains(api::SERVICE_ID) ) {
					return m_registry.get<api>();
				}

				// does not exist, invoke factory
				return buildAndStore<api>();
			}

			template<ServiceAPI api>
			void bind(ServiceBinding&& binding) {
				m_factories[ api::SERVICE_ID ] = binding;
			}

			template<ServiceAPI api>
			ServiceBindingT<api>& createBinding() {
				auto ptr = std::make_shared<ServiceBindingT<api>>(this);
				m_factories[ api::SERVICE_ID ] = ptr;
				return *(ptr.get());
			}

			template<ServiceAPI api>
			Service<api> build() {
				auto ptrRaw = invokeFactoryRaw(api::SERVICE_ID);
				return Service( std::static_pointer_cast<api>(ptrRaw) );
			}

			template<ServiceAPI api>
			std::optional<Service<api>> getFromCache() const {
				if ( !m_registry.contains(api::SERVICE_ID) ) {
					return {};
				}
				return m_registry.get<api>();
			}

			inline bool canBuild(ServiceName name) const {
				return m_registry.contains(name) || m_factories.find(name) != m_factories.end();
			}

		private:
			Registry m_registry;
			std::map<ServiceName, std::shared_ptr<ServiceBinding>> m_factories;

			std::shared_ptr<void> invokeFactoryRaw(ServiceName name);

			template<ServiceAPI api>
			Service<api> buildAndStore() {
				auto ptr = invokeFactoryRaw(FGGL_SERVICE(api));
				m_registry.provideUnsafe(FGGL_SERVICE(api), ptr);
				return m_registry.get<api>();
			}


	};

	template<ServiceAPI T>
	struct ServiceBindingT : public ServiceBinding {
			using Factory = std::function<std::shared_ptr<T>(Generator* generator)>;

			inline explicit ServiceBindingT(Generator* generator) : m_generator(generator) {}

			inline ServiceBindingT<T>& toFactory(Factory factory) {
				m_instance = nullptr;
				m_factory = factory;
				return *this;
			}

			inline ServiceBindingT<T>& asSingleton() {
				m_singleton = true;
				return *this;
			}

			inline ServiceBindingT<T>& simply() {
				m_instance = nullptr;
				m_factory = [](const Generator* /*generator*/) {
					return std::make_shared<T>();
				};
				return *this;
			}

			template<ServiceAPI... Deps>
			inline ServiceBindingT<T>& depends(){
				m_instance = nullptr;
				m_factory = [](Generator* generator) {
					return std::make_shared<T>( generator->getLazy<Deps>().ptr()... );
				};
				return *this;
			}

			template<Subclass<T> Impl>
			inline ServiceBindingT<T>& simplyTo(){
				m_instance = nullptr;
				m_factory = [](const Generator* /*generator*/) {
					return std::make_shared<Impl>();
				};
				return *this;
			}

			template<Subclass<T> Impl, ServiceAPI... Deps, typename... Args>
			inline ServiceBindingT<T>& to(Args... args){
				m_instance = nullptr;
				m_factory = [args...](Generator* generator) {
					return std::make_shared<Impl>( generator->getLazy<Deps>().ptr()..., args... );
				};
				return *this;
			}

			std::shared_ptr<void> provide() override {
				return provideT();
			}

			inline std::shared_ptr<T> provideT() {
				assert( m_generator != nullptr );
				if ( m_instance != nullptr ){
					return m_instance;
				}

				if ( m_factory != nullptr ) {
					if ( m_singleton ) {
						m_instance = m_factory(m_generator);
						return m_instance;
					}
					return m_factory(m_generator);
				}
				return nullptr;
			}

		private:
			Generator* m_generator = nullptr;
			std::shared_ptr<T> m_instance = nullptr;
			Factory m_factory = nullptr;
			bool m_singleton = false;
	};

	template<ServiceAPI impl, typename... Args>
	std::shared_ptr<void> wire_func(Generator* view) {
		return std::make_shared<impl>( view->getLazy<Args...>() );
	}

} // namespace fggl::services

#endif //FGGL_COMPONENTS_CORE_SERVICES_FACTORY_HPP
